#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <errno.h>
#include "parser.tab.h"
#include "apc.h"
#include <unistdio.h>
#include <unistr.h>
/* Public */
int  lexer_init(void);
void lexer_quit(void);
int  lexer_lexfile(uint8_t*);
int  lexer_lexdir(uint8_t*);
int  lexer_lexstring(uint8_t*, int);
//apc.c
static
yypstate* pstate;
static
yycstate* cstate;
/* Ring buffer for keeping lexical tokens valid for up to 255 tokens */
static
YYSTYPE   lval_stack[0xFF + 1];
static
uint8_t   lval_offs;
#define PUSHTOK(T,L) yypush_parse(pstate, T, (L), cstate)
#define LEXTOK(T,Y,L) do {			\
    lval_stack[lval_offs].Y = L;		\
    PUSHTOK(T,lval_stack + lval_offs);		\
    lval_offs++;				\
    ntok++;					\
  } while (0);
#define PUSHFACE(F) LEXTOK(FACING, face, F)
#define PUSHREF(R)  LEXTOK(REF, ref, R)
#define PUSHLINK()  LEXTOK(LINK, val, 0)
#define PUSHNUM(N)  LEXTOK(NUM, val, N)
#define PUSHNAME(N) LEXTOK(NAME, str, N)
#define PUSHOP(O)   LEXTOK(O, val, 0)
#define PUSHPATH(P) LEXTOK(PATH, str, P)

/* Lexstring is the main lexer for APC and is generated by ragel. It lexes file names of files 
   that have been scanned and pushes their types and values into the tok_stack, which yyparse 
   eventually calls during parsing. */

%%{    
  machine lexstring;

  # set up yylval and tok_t to be pushed to stack
  action push_ref       { errno = 0;
                          lval.ref = strtoll((char*)ts,NULL,16);
			  if (errno)
			    { fprintf(stderr, "Invalid hex number in file %s\n",(char*)str);
			      exit(1);
			    }
			  PUSHREF(lval.ref);
			}
  action push_link      { PUSHLINK(); }
  action push_val       { errno = 0;
                          lval.val = strtoll((char*)ts,NULL,10);
			  if (errno)
  			    { fprintf(stderr, "strtoll could not parse %s\n", (char*)str);
			      exit(1);
			    }
			  PUSHNUM(lval.val);
                        }
  action push_name      { //printf("Lexer_lexstring:: action:push_name: from %s to %s\n", ts, p);
			  PUSHNAME(ts);
                        }
  action push_map       { //printf("Lexer_lexstring:: action:push_map: pushing map token\n");
			  PUSHOP(MAP);
		        }
  action set_ts         { //printf("Lexer_lexstring:: action:set_ts. ts = %s\n", p);
                          ts = p; }
  action push_SS        { //printf("Lexer_lexstring:: action:push_SS. p = %s\n",p);
                          PUSHOP(SS);
                        }
  action push_S         { //printf("Lexer_lexstring:: action:push_S. p = %s\n", p);
			  PUSHFACE(SFACE);
                        }
  action push_SW        { //printf("Lexer_lexstring:: action:push_SW. p = %s\n", p);
                          PUSHFACE(SWFACE);
		        }
  action push_W         { //printf("Lexer_lexstring:: action:push_W. p = %s\n", p);
                          PUSHFACE(WFACE);
                        }
  action push_NW        { //printf("Lexer_lexstring:: action:push_NW. p = %s\n", p);
			  PUSHFACE(NWFACE);
                        }
  action push_N         { //printf("Lexer_lexstring:: action:push_N. p = %s\n", p);
                          PUSHFACE(NFACE);
                        }
  action push_NE        { //printf("Lexer_lexstring:: action:push_NE. p = %s\n", p);
			  PUSHFACE(NEFACE);
                        }
  action push_E         { //printf("Lexer_lexstring:: action:push_N. p = %s\n", p);
			  PUSHFACE(EFACE);
                        }
  action push_SE        { //printf("Lexer_lexstring:: action:push_N. p = %s\n", p);
			  PUSHFACE(SEFACE);
                        }
  action ref_error      { //printf("ref from %s to %s has an inappropriate amount of hex digits, it must have eight.\n", ts, p);
                          exit(1);
                        }
  action p              { //printf("Lexer_lexstring:: p = %s\n", p);
                        }
    
  N = 'N' %push_N;
  W = 'W' %push_W;
  S = 'S' %push_S;
  E = 'E' %push_E;
  NW = 'NW' %push_NW;
  NE = 'NE' %push_NW;
  SW = 'SW' %push_SW;
  SE = 'SE' %push_SE;

  tok_delimiter = [_];

  direction = (N | W | S | E | NW | NE | SW | SE) ;
  dimensions = (digit+ - '0') >set_ts %push_val 'x' (digit+ - '0') >set_ts  %push_val;
  link = '#'  %push_link;
  SS = ('+SS' %to(push_SS)) | ('+SS' %to(push_SS) link ) ;
  ref = '0x' >set_ts alnum{8} $err(ref_error) %push_ref ;
  val = digit+ >set_ts %push_val ;
  name = lower >set_ts (lower | digit)*  %push_name ;
  map = '+MAP' %to(push_map);
  tok = (name | val | ref | dimensions | map | link | SS | direction);
  

  main := (tok tok_delimiter)* tok [\0];

  write data nofinal noerror noprefix;
  
}%%

int lexer_init
( void )
{ pstate = yypstate_new();
  cstate = yycstate_new();
  lval_offs = 0;
  return !pstate || !cstate;
}

void lexer_quit
( void )
{ if (pstate) yypstate_delete(pstate);
  if (cstate) yycstate_delete(cstate);
}

int lexer_lexstring
( uint8_t* str, 
  int size
)
{ uint8_t *p; 
  uint8_t *ts, *pe, *eof;
  int cs, ntok;
  YYSTYPE lval;

  ntok = 0;
  p = ts = str;
  pe = eof =  p + size + 1;

  //printf("|---Begin lexstring on p = %s, pe = %s.\n",p, pe);

  %%write init;
  %%write exec;

  //printf("Ending lexstring of file %s, pushed %d tokens.\n",str, ntok);

  return ntok;
}

/* Lexical analysis of a file
   Strips a filename to its base name, then sends it to lexer_lexstring before
   pushing a PATH token with the filename
   Returns the number of tokens pushed to the parser.
*/
int lexer_lexfile
( uint8_t* filename )
{ uint8_t* last_period,* iter,* filename_end;
  int      ntok;
  last_period = NULL;
  for (iter = filename; *iter; iter++)
    switch (*iter)
      { // Keep track of the last 'dot' in the name
        case '.' : last_period = iter; continue;
        // replace '_' with '\0' so bison can use strlen on them as tokens.
        case '_' : *iter = '\0';
        default: continue;
      }
  // Mark the end of the filename
  filename_end = iter;
  // Lex from either the last period, if present, or filename end
  ntok = (last_period) ?
    lexer_lexstring(filename, (int)(last_period - filename))
  : lexer_lexstring(filename, (int)(iter - filename));
  // Replace nulls with their original '_'
  for (iter = filename; iter < filename_end; iter++)
    if (*iter == '\0')
      *iter = '_';
  PUSHPATH(filename);
  return ntok + 1;
  return en_main == 1;
}

int lexer_lexdir
( uint8_t* dirname )
{ uint8_t* de = dirname;
  int ntok;
  ntok = 0;
  de = dirname;
  if (*de) while (*++de);
  ntok = lexer_lexstring(dirname, (int)(de - dirname));
  PUSHOP(CLOPEN);
  return ntok;
}

int lexer_closedir
( void )
{ int ntok = 0;
  PUSHOP(CLCLOSE);
  return ntok;
}
